open Ast
open System.IO
open Typecheck
open Print

(*types*)
type Lemma = string
type Proof = ((( Clause * Lemma list) option) Set.t)
type ValueTable = (Term list, Lemma list) Hashtbl.t
type ArrangedTable = (Term list, ((Term list)*(Lemma list)) list) Hashtbl.t
type NamedTable = ValueTable * (Term list);;
(* this is an intermediate data structure used to do joins
   it is a pair, where the first element is a ValueTable
   and the second element is a list of columnnames for the
   tuples, which are keys in the value table *)

(*global variables*)
let proofTable : ( (Lemma,Proof) Hashtbl.t) = Hashtbl.create 17;;
let db: (Predicate, ValueTable ) Hashtbl.t  = Hashtbl.create 17;;
let dbChanged = ref true;;

let findOrCreateInDb pred =
	try
	    Hashtbl.find db pred 
	with
	    Not_found -> let newHash = Hashtbl.create 5 in
		        Hashtbl.add db pred newHash;
			newHash
;;

let getPredicateFromLiteral (lit:Literal) =
    let (pred,_) = lit in
    pred
;;
let getStringFromLiteral (lit:Literal) =
    let (Pred(s),_) = lit in
    s
;;
(* build dependency graph *)
let buildDependency (rl:Clause list) : ( ( (string) Set.t) *  ( (bool*string*string) Set.t))=
	let vertices = ref (Set.empty) and edges = ref (Set.empty) 
	
	in let addVertex (v:string) : unit =
		if Set.mem v !vertices 
		then () 
		else vertices := Set.add v !vertices 
	
	in let addEdge (u:string) (v:string) (p:bool) : unit= 
		let e= (p, u, v) in
		if Set.mem e !edges then ()
		else edges := Set.add e !edges
		
        in let addPositiveDependency fromVertex toVertex =
	    addEdge fromVertex toVertex true

	in let addNegativeDependency fromVertex toVertex =
	    addEdge fromVertex toVertex false
	
	in let processOneRule rule =
		match rule with
		| Rule(lhsLiteral, rhsPositiveLiterals, rhsNegativeLiterals, filters, functions) ->
		    let lhs = getStringFromLiteral lhsLiteral in
 		    addVertex lhs;
		    List.iter (addPositiveDependency lhs) (List.map getStringFromLiteral rhsPositiveLiterals);
		    List.iter (addNegativeDependency lhs) (List.map getStringFromLiteral rhsNegativeLiterals)
		    
	        |_ -> (assert(false))	(* this should never happen *)
		
	in List.iter (fun y -> processOneRule y) rl; 
	(!vertices, !edges)
;;


(* takes a DAG G=(V,E) and a root node as input *)
let revTopSort vertices edges root =
	let nr = ref Set.empty in 
	(vertices, edges)
;;

let adjL v e = 
	Set.fold (fun x y -> match x with
				| (_,a,b) -> if (String.compare a v)=0 then b::y else y) e [] 
;;

let revG v e = 
	Set.fold (fun x y -> match x with 
				| (p,a,b) -> Set.add (p,b,a) y) e (Set.empty)
;;


let rec dfsVisit x s e (ll: string list ref)  =
	s := Set.add x !s;
	let l = adjL x e  in
		List.iter (fun z ->  if (Set.mem z !s) then () else  ignore(dfsVisit z s e ll)
			  ) l;
        ll := x::!ll;
	!ll
;;



let dfs v e =
	let s = ref (Set.empty) and ll = ref [] in
		Set.iter (fun x -> if (Set.mem x !s) then () else ignore(dfsVisit x s e ll)) v;
		!ll

;;

let dfs2 v e vl =
	let s = ref (Set.empty) in
		let sccs = List.map (fun x -> if (Set.mem x !s) then [] else dfsVisit x s e (ref [])) vl 
                in
		let nonNullSccs = List.filter (fun x -> match x with
					| [] -> false
					| _ -> true)	sccs in
		let revNonNullSccs = List.rev nonNullSccs in
		revNonNullSccs
;;

let printOneScc scc =
       printf "[";
       List.iter (fun x -> printf "%s" x) scc;
       printf "]"
;;

let printSccs sccs : unit =
	 List.iter printOneScc sccs; printf "\n"
;;
// visits vertices according to finish times
let sccs v e =
	let l = dfs v e in
	dfs2 v (revG v e) l
;;


    
let rec matchCommonColumnsAux (t1:Term list) (t2:Term list) (index:int)  :  ((Term*int*int) list) =
    let rec matchOneElement (t:Term) (terms:Term list) =
	match terms with
	    | [] -> None
	    | hd::tl -> if(hd = t) then Some(0)
		        else
			  let temp = matchOneElement t tl in
			  match temp with
			     | None -> None
			     | Some(x) -> Some(x+1)
  in
  match t1 with
      | [] -> []
      | hd::tl ->
	  let matchResult = matchOneElement hd t2
	  in
	  match matchResult with
	      | Some(x) -> (hd,index,x) :: matchCommonColumnsAux tl t2 (index+1)
	      | None    -> matchCommonColumnsAux tl t2 (index+1)
;;

let matchCommonColumns t1 t2 =
    matchCommonColumnsAux t1 t2 0
;;

let matchOnlyInTable2 (t1:Term list) (t2:Term list) : (Term list) =
    let myFilter a =
	   (not (List.exists (fun x -> (x=a)) t1)) in
    List.filter myFilter t2
;;	      

(* from x::y::z make a hash table saying (x->0,y->1,z->2) *)
let getHeadingsHash headings =
    let newHash = Hashtbl.create 5 in
    let _ = List.fold_left
	( fun index term -> Hashtbl.add newHash term index; index+1)
    0 headings
    in
      newHash
;;

let select (valueTable:ValueTable) (values:(int*Term)list) : ValueTable = 
    (* assume that values is sorted by index *)
    let (newTable:ValueTable) = Hashtbl.create 17 in
    let matchOneTuple terms lemmas  =
	(* match all the values with the terms *)
	let rec matchOneTupleAux terms v resultsSoFar index : ((Term list) option)   =
	    if(v = []) then Some(List.append resultsSoFar terms)
	    else
		match terms with
		    | [] -> Some(resultsSoFar)
		    | hd::tl ->
			(* List.hd and List.tl should never fail here *)
			let (i,t) = List.hd v in
			if( i = index) then
			    if(hd = t) then
				matchOneTupleAux tl (List.tl v) resultsSoFar (index+1)
			    else
				None
			else if(i < index) then
				begin (assert(false)); None end
				(* this should never happen *)
			else (* i > index *)
			    matchOneTupleAux tl v (List.append resultsSoFar [hd]) (index+1)
		    
	in
	let result = matchOneTupleAux terms values [] 0 in
	match result with
	    |Some(t) -> Hashtbl.add newTable t lemmas
	    |_ -> ()
    in
    Hashtbl.iter matchOneTuple  valueTable;
    newTable 
;;

let project (valueTable:ValueTable) (columns: int list): ValueTable =
    let (newTable:ValueTable) = Hashtbl.create 17 in
    let matchOneTuple terms lemmas =
	let newTuple = List.map  (fun index -> List.nth terms index)   columns
	in Hashtbl.add newTable newTuple lemmas
    in
    Hashtbl.iter matchOneTuple valueTable;
    newTable
;;

(* obsolete, has been replaced by newJoin *)
let join (namedTable1:NamedTable) (namedTable2:NamedTable) : NamedTable =
    let (table1, heading1)= namedTable1 in
    let (table2, heading2)= namedTable2 in 
    let matches:( (Term*int*int) list) = matchCommonColumns heading1 heading2 in
    let sortedMatches = List.sort (fun (_,_,a) (_,_,b) -> (a - b))  matches in
    let onlyInTable2 = matchOnlyInTable2 heading1 heading2 in
    let newTable = Hashtbl.create 17 in 
    Hashtbl.iter ( fun tuple lemmas ->
	           let values:( (int*Term) list)  = List.map (fun (_,i,j) -> (j,(List.nth tuple i))) sortedMatches
	           in
		   let subTable = select table2 values in
		   Hashtbl.iter (fun tuple2 lemmas2 -> Hashtbl.add newTable (List.append tuple tuple2) (List.append lemmas lemmas2)) subTable
		 )
		 table1;
    (newTable, (List.append heading1 onlyInTable2))
;;    
   
let arrangeTable (table:ValueTable) (columns:(Term*int*int) list) : ArrangedTable =
    let (arrTable:ArrangedTable) = Hashtbl.create 17 in
    Hashtbl.iter (
	fun (tuple:Term list) (lemmas:Lemma list) ->

	(* extract common columns from first table *)
	let commonColumns = List.fold_right (
		fun column tupleSoFar ->
			let (term,index,_) = column in
			(List.nth tuple index)::tupleSoFar
	) columns [] in

	try
		let currentEntries = Hashtbl.find arrTable commonColumns in
		let allEntries = (tuple,lemmas)::currentEntries in
		Hashtbl.replace arrTable commonColumns allEntries
	with Not_found ->
		Hashtbl.add arrTable commonColumns [(tuple,lemmas)]

    ) table;
    arrTable
;;

let newJoin (namedTable1:NamedTable) (namedTable2:NamedTable) ignorelist : NamedTable =
    let (table1,heading1) = namedTable1 in
    let (table2,heading2) = namedTable2 in 
    let matches:((Term*int*int) list) = matchCommonColumns heading1 heading2 in
    let arrangedTable1 = arrangeTable table1 matches in

    let onlyInTable2 = matchOnlyInTable2 heading1 heading2 in
    let hdgHash2 = getHeadingsHash heading2 in
    (*let hdgHash = getHeadingsHash (List.append heading1 onlyInTable2) in
    *)
    let (newHash:ValueTable) = Hashtbl.create 17 in

	let rec ilist xs n = match xs with
		| [] -> ([],[])
		| x::rest -> let (ns,xs) = ilist rest (n+1) in 
		if List.mem x ignorelist then (ns,xs) else (n::ns,x::xs) in
	
	let listvars = List.append heading1 onlyInTable2 in
	let (indexlist,varlist) = ilist listvars 0 in
	


    Hashtbl.iter (
	fun tuple2 lemmas2 ->
	
	(* extract common columns from second table *)
	let commonColumns = List.fold_right (
		fun column tupleSoFar ->
			let (term,_,index) = column in
			(List.nth tuple2 index)::tupleSoFar
	) matches [] in

	(* extract columns only occurring in the second table *)
	(* Hashtbl.find should never fail *)
	let columnsOnlyInTable2 = List.map (
		fun x -> (List.nth tuple2 (Hashtbl.find hdgHash2 x))
	) onlyInTable2 in
(*
	let indexlist = List.map (Hashtbl.find hdgHash) (Set.to_list ignoreset) in
*)

	
	try
		let currentEntries = Hashtbl.find arrangedTable1 commonColumns in
		List.iter (
			fun (tuple1,lemmas1) ->
			Hashtbl.add newHash (List.append tuple1 columnsOnlyInTable2) (List.append lemmas1 lemmas2)
		) currentEntries
	with Not_found -> ()
    ) table2;
    (project newHash indexlist,varlist)
;;

let doPositiveJoin (namedTableListOpt: ((NamedTable*(Collections.Set<Term>)) list)) : NamedTable =
    match namedTableListOpt with
	| (head,y)::tail ->
		let ignorelist = Set.to_list y in  
		let (hvalt,listvars) = head in
        let rec ilist xs n = match xs with
			| [] -> ([],[])
			| x::rest -> let (ns,xs) = ilist rest (n+1) in 
				if List.mem x ignorelist then (ns,xs) else (n::ns,x::xs) in
		let (indexlist,varlist) = ilist listvars 0 in
		
		List.fold_left (fun curResult (x,y) -> newJoin curResult x (Set.to_list y)) (project hvalt indexlist,varlist) tail
	| [] -> assert(false); (Hashtbl.create 17,[])
		(* This should never fail because of restrictions imposed by the parser. *)
;;    

let doNegativeJoin negJoinConditions (currentTable: NamedTable) : NamedTable =
    let (valTable,currHeadings) = currentTable in
    let hdgHash = getHeadingsHash currHeadings in

    let newTable = Hashtbl.create 17 in
    Hashtbl.iter
	(fun tableTuples lemmas ->

	let testOneNegJoin negJoin : bool =
		let (pred,terms) = negJoin in
		try
		    let negativeDB = Hashtbl.find db pred in
		    let negativeTuple = List.map (
			fun term -> match term with
			    (* Hashtbl.find should never fail here due to sanity checks. *)
			    | Var(_) -> (List.nth tableTuples (Hashtbl.find hdgHash term))
			    | _ -> term
			) terms
		    in
		    let _ = Hashtbl.find negativeDB negativeTuple in false
		with Not_found -> true
	in

	let rec testFilterConditions remainingFilters =
		match remainingFilters with
			| [] -> true;
			| filter::rest -> if (not (testOneNegJoin filter)) then false
					  else testFilterConditions rest
	in

	if testFilterConditions negJoinConditions then 
	    	Hashtbl.add newTable tableTuples lemmas
	else ()
	)
    valTable;
    (newTable,currHeadings)
;;

let filterConditions filters currentTable =
    let (valTable,currHeadings) = currentTable in
    let hdgHash = getHeadingsHash currHeadings in

    let newTable = Hashtbl.create 17 in
    Hashtbl.iter
	(fun tableTuples lemmas ->

	let getTermValue term =
		match term with
		(* Hashtbl.find should never fail here due to sanity checks. *)
		| Var(_) -> (List.nth tableTuples (Hashtbl.find hdgHash term))
		| _ -> term
	in
	let testOneCondition filter =
	    match filter with
	    | LessThan(term1,term2) ->
		   ((getTermValue term1) < (getTermValue term2))
	    | GreaterThan(term1,term2) ->
		   ((getTermValue term1) > (getTermValue term2))
	    | LessThanEqual(term1,term2) ->
		   ((getTermValue term1) <= (getTermValue term2))
	    | FilterEqual(term1,term2) ->
		   ((getTermValue term1) = (getTermValue term2))		  
	in
	let rec testFilterConditions remainingFilters =
		match remainingFilters with
			| [] -> true;
			| filter::rest -> if (not (testOneCondition filter)) then false
					  else testFilterConditions rest
	in

	if testFilterConditions filters then 
	    	Hashtbl.add newTable tableTuples lemmas
	else ()
	)
    valTable;
    (newTable,currHeadings)
;;

let extendWithFunctionValues functions (currentTable:NamedTable) : NamedTable =
    let (valTable,currHeadings) = currentTable in
    let hdgHash = getHeadingsHash currHeadings in
    let extraHeadings = List.map (fun fn -> match fn with | FnEqual(x,_) -> x) functions in
    let (newTable:ValueTable) = Hashtbl.create 17 in
    Hashtbl.iter
	(fun (tableTuples:Term list) (lemmas:Lemma list) ->
	    let rec computeExpr (expr:Expression) : int =
		match expr with
		| Plus(e1,e2) -> (computeExpr e1) + (computeExpr e2)
		| Minus(e1,e2) -> (computeExpr e1) - (computeExpr e2)
		| Terminal(t) -> 
			match t with
			| Int(n) -> n
			| String(_) -> (assert(false)); 0
			  (* Hashtbl.find should never fail here due to sanity checks. *)
			| Var(_) -> match (List.nth tableTuples (Hashtbl.find hdgHash t)) with
				  | Int(n) -> n
				  | _ -> (assert(false)); 0
				  (* type checking assures that this assert never triggers *)
	    in
	    let computeFn f : Term =
		    match f with FnEqual(_,expr) -> 
			match expr with
			| Terminal(t) -> (match t with
				| String(s) -> t
				| _ -> Int(computeExpr expr))
			| _ -> Int(computeExpr expr)
	    in
	    let (extraFields:Term list) = List.map computeFn functions in
	    Hashtbl.add newTable (List.append tableTuples extraFields) lemmas
	)
    valTable;
    (newTable,(List.append currHeadings extraHeadings))
;;

let convertLiteralToNamedTable literal =
    let (pred,terms) = literal in
    let (constantList,termList,_) = 
                   List.fold_right
		       (
                        fun cur (constsSoFar,termsSoFar, index)  ->
			   match cur with
			       | Var(_) -> (constsSoFar, (cur::termsSoFar), index-1)
			       | Int(_) -> ( ((index,cur)::constsSoFar) , termsSoFar, index-1)
			       | String(_) -> ( ((index,cur)::constsSoFar) , termsSoFar, index-1)
		       )
		       terms
		       ([],[], (List.length terms)-1)
    in
    let valueTable = 
	try Hashtbl.find db pred
	with Not_found -> Hashtbl.create 5
    in
    ((select valueTable constantList),termList)
;;		       

let returnIndex list elem : int =
	let rec returnIndexAux list index : int =
	match list with
	   | [] -> (assert(false)); 0
	   | head::tail -> if (head=elem) then index
			   else returnIndexAux tail (index+1)
	in
	returnIndexAux list 0
;;


let lookupProof lemmaString = 
	try
		Hashtbl.find proofTable  lemmaString
	with
		Not_found -> let nullProof = Set.empty in 
			     Hashtbl.add proofTable lemmaString nullProof;
			     dbChanged := true;
			     nullProof
;;

let applyOneRule rule =
	printf "applying %s\n" (getRuleName rule);
	match rule with
	| Rule(lhsLiteral, rhsPositiveLiterals, rhsNegativeLiterals, filters, functions) ->

   let filterVarsFromTuple tuple = List.filter (fun x -> match x with | Var(x) -> true | _ -> false) tuple in
   let varsInLit lit = match lit with | (_,tuple) -> Set.of_list (filterVarsFromTuple tuple) in
   let fnRhsVars fn =
	let rec fnRhsVarsAux e = match e with
		| Minus(e1,e2) -> Set.union (fnRhsVarsAux e1) (fnRhsVarsAux e2)
		| Plus(e1,e2) -> Set.union (fnRhsVarsAux e1) (fnRhsVarsAux e2)
		| Terminal(Var(x)) -> Set.singleton (Var(x))
		| _ -> Set.empty
	in match fn with | FnEqual(_,e) -> fnRhsVarsAux e in
   let filterVars fl = match fl with
	| LessThan(t1,t2)      -> Set.of_list (filterVarsFromTuple (t1::[t2]))
	| GreaterThan(t1,t2)   -> Set.of_list (filterVarsFromTuple (t1::[t2]))
	| LessThanEqual(t1,t2) -> Set.of_list (filterVarsFromTuple (t1::[t2]))
	| FilterEqual(t1,t2)   -> Set.of_list (filterVarsFromTuple (t1::[t2])) in

   let allOtherVars = 			
   (List.fold_right Set.union (List.map filterVars filters) 
   (List.fold_right Set.union (List.map fnRhsVars functions) 
   (List.fold_right Set.union (List.map varsInLit rhsNegativeLiterals) 
   (varsInLit lhsLiteral)))) in

	let rec passToJoin litlist vars = match litlist with 
		| [] -> []
		| lit::rest -> let ntable = convertLiteralToNamedTable lit in
			let vars' = List.fold_right Set.union (List.map varsInLit rest) allOtherVars in
			(ntable,Set.diff vars vars')::(passToJoin rest vars') in
	 
	 let rhsPositiveNamedTablesOpt = passToJoin rhsPositiveLiterals 
		(List.fold_right Set.union (List.map varsInLit rhsPositiveLiterals) allOtherVars) in	

(*   let posVars = List.fold_right Set.union (List.map varsInLit rhsPositiveLiterals) Set.empty in
*)
	    (* Positive literals: get rhsPositiveLiterals from the database and convert them to appropriate named tables *)
(*	    let rhsPositiveNamedTables = List.map convertLiteralToNamedTable rhsPositiveLiterals in
*)
	    let filteredTable = filterConditions filters 
				(doNegativeJoin rhsNegativeLiterals
				(extendWithFunctionValues functions
				(doPositiveJoin rhsPositiveNamedTablesOpt))) in

	    let (values,terms) = filteredTable in
	    let (lhsPred, lhsTerms) = lhsLiteral in
	    let indices = List.map (returnIndex terms) lhsTerms in 
	    (* Hashtbl.find in returnIndex should never fail due to sanity checks. *)

	    let resultValueTable =  project values indices in
	    let dbResultTable = findOrCreateInDb lhsPred in
	    Hashtbl.iter
	        (fun tuples lemmas ->
			  let curLemmaString =  getLemmaString lhsPred tuples in
			  let(curProofs:Proof) = lookupProof  curLemmaString in
			  let (newProof: (Clause*Lemma list) option) = Some(rule, lemmas) in
			  if (not (Set.exists ( fun e -> (e=newProof)) curProofs)) then
				begin
				dbChanged := true;
	  		        let allProofs = Set.add newProof curProofs in
				Hashtbl.replace proofTable curLemmaString allProofs
				end;
			  let result =  
				try 
					Hashtbl.find dbResultTable tuples 
				with 
					Not_found -> 
					  begin
					   Hashtbl.add dbResultTable tuples [curLemmaString];
					   dbChanged := true;
					   [curLemmaString]
					  end
			  in
			  match result with
			     (* these checks should never fail *)
			  | [storedLemmaString] ->  (assert (storedLemmaString = curLemmaString))
			  | _ -> (assert(false))
                )
		resultValueTable
	|_ -> (assert(false))	(* this should never happen *)
;;

let rec applyRulesOnce ruleList =
     List.iter applyOneRule ruleList;
;;

let rec applyRulesFixPoint ruleList =
     dbChanged := false;
     applyRulesOnce ruleList;
     if(!dbChanged) then
	applyRulesFixPoint ruleList
;;


let addTupleHash (tbl:ValueTable) (terms:Term list) (lemmas:Lemma list) : unit =
     let _ = 
       try
	   Hashtbl.find tbl terms
       with
	   Not_found -> Hashtbl.add tbl terms lemmas;[]
     in ()
;;

let addClauseToDb fact =
    match fact with
    | Fact(pred,termList) ->
	(let (tupleHashTable:ValueTable) = findOrCreateInDb pred in
	let currLemmaString = getLemmaString pred termList in
        try
	    let _ = Hashtbl.find tupleHashTable termList in ()
	with Not_found ->
	    addTupleHash tupleHashTable termList [currLemmaString];
	    let mySet = Set.add None Set.empty in
	    Hashtbl.add proofTable currLemmaString mySet)
    | _ -> (assert(false))	(* this should never happen *)
;;

let isInScc var scc =
    List.exists (fun y -> y=var) scc
;;

let decomposeRulesIntoSccs rl =	   
    let (v,e) = buildDependency rl in
    let mySccs = sccs v e in

    let filterRulesByPredicateStringList (predStringList:string list) =
		List.filter 
			( fun x -> 
				match x with  
				| Rule((Pred(s),_),_,_,_,_) -> isInScc s predStringList
			 	| _ -> false
			)
			rl
    in

    let isRecursiveScc (predStringList:string list) =
	Set.exists (fun (_,u,v) ->
		(isInScc u predStringList) && (isInScc v predStringList)
	) e
    in

    List.map (
	fun scc -> (scc,filterRulesByPredicateStringList scc,isRecursiveScc scc)
    ) mySccs
;;

let checkStratified (scc,rules,_) =
    List.iter (
	fun rule -> match rule with
	    | Rule(_,_,negList,_,_) -> 
		List.iter (fun negPred -> match negPred with | (Pred(x),_) -> 
		    if isInScc x scc then begin
			printf "ERROR: non-stratified use of negated predicate %s; aborting\n" x;
			exit 1
		    end
		) negList
	    | _ -> (assert(false))	(* this should never fail *)
    ) rules
;;

let saturate (fl: Clause list) (rl: Clause list) =
    List.iter addClauseToDb fl;
    List.iter typeCheckFact fl;
    let rulesListList = decomposeRulesIntoSccs rl in
    List.iter checkStratified rulesListList;
    List.iter (fun (_,rules,_) -> List.iter typeCheckRule rules) rulesListList;
    List.iter (fun (_,rules,b) -> 
	if b then
		applyRulesFixPoint rules
	else
		applyRulesOnce rules
    ) rulesListList
;;
   
