open System.IO
open Ast
open Print
open Ninterp

(* This function prints all (initial or derived) facts found for a given predicate. *)
let printPredTable pred valTable =
	Hashtbl.iter
	(fun tuples lemmas ->
		printf "%s <- " (getLemmaString pred tuples);
		match lemmas with
				(* Hashtbl.find should never fail here *)
		| [lem] -> let proof = Hashtbl.find proofTable lem in
			Set.iter ( fun elem ->
				match elem with
				| Some(_,list) ->
					List.iter (fun t -> printf "%s " t) list;
					printf "; ";
				| None -> printf "initial ; "
			) proof;
			printf "\n";
		| _ -> (assert(false))	(* this should never happen *)
	)
	valTable
;;

(* This function prints all (initial or derived) facts found for a given query. *)
let printQueryTable query (stream:StreamWriter) =
	let proof = Hashtbl.find proofTable query in
			Set.iter ( fun elem ->
				match elem with
				| Some(_,list) ->
					List.iter (fun t -> stream.Write(sprintf "%s " t)) list;
					stream.Write("\n");
				| None -> ()
			) proof
;;

(* This function prints all (initial and derived) facts for all predicates. *)
let printAllTables x =	(* dummy argument, to prevent this from executing by itself *)
	Hashtbl.iter printPredTable db
;;

let xmlPurify (s:string) : string =
    let t = ref "" in
     String.iter (fun c ->
	if c = '"' then
	    t := sprintf "%s&quot;" !t 
	else if c = '<' then
	    t := sprintf "%s&lt;" !t 
	else if c = '>' then
	    t := sprintf "%s&gt;" !t 
	else if c = '&' then
	    t := sprintf "%s&amp;" !t 
	else if c = '(' then
	    t := sprintf "%s (" !t 
	else if c = ',' then
	    t := sprintf "%s, " !t 
        else
	    t := sprintf "%s%c" !t c
     ) s;
     !t
;;

(* This prints the proof for a given lemma. *)

let printProof (lemma:Lemma) (stream:StreamWriter) printAll =

    let (visitedNodes : (Lemma,unit) Hashtbl.t) = Hashtbl.create 17 in

    let printProofPrologue lemma =
	stream.WriteLine("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
	stream.WriteLine("<graphml xmlns:y=\"http://yworks.com/xml/graphml\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://graphml.graphdrawing.org/xmlns/graphml http://www.yworks.com/xml/schema/graphml/1.0/ygraphml.xsd\">");
	stream.WriteLine("  <key id=\"d0\" for=\"node\" yfiles.type=\"nodegraphics\"/>");
	stream.WriteLine("  <key id=\"d1\" for=\"edge\" yfiles.type=\"edgegraphics\"/>");
	stream.WriteLine("  <graph id=\"G\" edgedefault=\"directed\">");
    in

    let printProofEpilogue lemma =
	stream.WriteLine("  </graph>\n</graphml>")
    in

    let printProofNode lemma =
	let nodeSize name =
		let length = String.length name in
		if length > 12 then 100 else length*7
	in

	let nodeColour lemma : string =
	    if (((String.length lemma) > 5) && ((String.sub lemma 0 5) = "Query")) then
		"#FF6600"
	    else
		(* Hashtbl.find should never fail here *)
		let proof = Hashtbl.find proofTable lemma in
		Set.fold ( fun branch colour ->
			match branch with
			| Some(_,_) -> colour
			| None -> "#FF0000"
		) proof "#FFCC00"
	in

	stream.WriteLine(sprintf "    <node id=\"%s\"><data key=\"d0\"><y:ShapeNode>" (xmlPurify lemma));
	stream.WriteLine(sprintf "       <y:Geometry width=\"%d.0\"/>" (nodeSize lemma));
	stream.WriteLine(sprintf "       <y:NodeLabel>%s</y:NodeLabel>" (xmlPurify lemma));
	stream.WriteLine(sprintf "       <y:Fill color=\"%s\"/>" (nodeColour lemma));
        stream.WriteLine("    </y:ShapeNode></data></node>")
    in

    let proofGenHnode lemma rule index =
	let nodeName = sprintf "%s_%d" lemma index in
	stream.WriteLine(sprintf "    <node id=\"%s\"><data key=\"d0\"><y:ShapeNode>" (xmlPurify nodeName));
	stream.WriteLine("       <y:Geometry width=\"10.0\" height=\"10.0\"/>");
	stream.WriteLine(sprintf "       <y:NodeLabel visible=\"false\">%s</y:NodeLabel>" (xmlPurify (getRuleName rule)));
	stream.WriteLine("       <y:Fill color=\"#00CCFF\"/>");
        stream.WriteLine("    </y:ShapeNode></data></node>");
	nodeName
    in

    let printProofEdge node1 node2 =
	let p1 = xmlPurify node1 in
	let p2 = xmlPurify node2 in
	stream.WriteLine(sprintf "    <edge id=\"%s/%s\" source=\"%s\" target=\"%s\"></edge>" p1 p2 p1 p2)
    in

    let rec proofDfs lemma printAll =

	let printOneProof oneProof =
		(match oneProof with
		| Some(rule,list) ->
			let hypNode = proofGenHnode lemma rule 1 in
			printProofEdge lemma hypNode;
			List.iter (fun t -> 
				proofDfs t false;
				printProofEdge hypNode t
			) list
		| None -> ())
	in

	try
		Hashtbl.find visitedNodes lemma
	with Not_found ->
	Hashtbl.add visitedNodes lemma ();
	printProofNode lemma;

	(* Hashtbl.find should never fail here *)
	let proof = Hashtbl.find proofTable lemma in

	if printAll then begin

	    Set.iter printOneProof proof

	end else begin

	    printOneProof (Set.choose proof)

	end

    in

    printProofPrologue lemma;
    proofDfs lemma printAll;
    printProofEpilogue lemma
;;

(* Printed out all tables to "dumpall".
let printQueryResults x = printAllTables x
;;
*)


let printQueryResultsforXRay (ql:Clause list) (flag:string) =
	let printAll = flag.Equals("1") in
     printf "\n";
     let filename2 = String.lowercase (sprintf "AllResults.txt") in
     printf "printing result to  %s\n" filename2;  
		//let filename = String.lowercase (sprintf "Attacks\%s.graphml" x) in
    let stream2 = new StreamWriter(filename2) in
    List.iter (fun cl ->
	match cl with
	|Rule((Pred(x),[]),[rhs],_,_,_) -> (
	    let (lemma:string) = getLemmaString (Pred(x)) [] in
	    try let _ = Hashtbl.find proofTable lemma in
		    printQueryTable lemma stream2;
	    //(try
		    //  let stream = new StreamWriter(filename) in
		    // printProof lemma stream printAll;
		    //stream.Close()
		// with _ -> printf "ERROR writing to file %s; proof not written to file\n" filename)
	    with Not_found ->
		printf "could not prove %s\n" (getLiteralString rhs)
	  )
	| _ -> (assert(false))	(* this should never happen *)
    ) ql;
    stream2.Close()  
;;

let printQueryResults (ql:Clause list) (flag:string) =
	let printAll = flag.Equals("1") in
    printf "\n";
    List.iter (fun cl ->
	match cl with
	|Rule((Pred(x),[]),[rhs],_,_,_) -> (
	    let (lemma:string) = getLemmaString (Pred(x)) [] in
	    try let _ = Hashtbl.find proofTable lemma in
		let filename2 = String.lowercase (sprintf "Attacks\%s.txt" x) in
		let filename = String.lowercase (sprintf "Attacks\%s.graphml" x) in
		printf "printing result to %s and %s\n" filename2 filename;
		(try
		    let stream2 = new StreamWriter(filename2) in
		    printQueryTable lemma stream2;
		    stream2.Close()
		with _ -> printf "ERROR writing to file %s; proof not written to file\n" filename);
	    (try
		    let stream = new StreamWriter(filename) in
		    printProof lemma stream printAll;
		    stream.Close()
		with _ -> printf "ERROR writing to file %s; proof not written to file\n" filename)
	    with Not_found ->
		printf "could not prove %s\n" (getLiteralString rhs)
	  )
	| _ -> (assert(false))	(* this should never happen *)
    ) ql
;;