open Ast
open Print

let typeTable : (Predicate,Term list) Hashtbl.t = Hashtbl.create 17;;

let errNoVars where =
	printf "Error in %s: variables not allowed in facts. Aborting.\n" where;
	exit 1
;;

let errIntWhereString where =
	printf "Error in %s: found integer, expecting string. Aborting.\n" where;
	exit 1
;;

let errStringWhereInt where =
	printf "Error in %s: found string, expecting integer. Aborting.\n" where;
	exit 1
;;

let errArity where =
	printf "Error in %s: does not have the expected arity. Aborting.\n" where;
	exit 1
;;

let errMixup where =
	printf "Error in %s: variable used with both string and integer fields. Aborting.\n" where;
	exit 1
;;

let errComparison where =
	printf "Error in %s: comparison between string and integer. Aborting.\n" where;
	exit 1
;;
let createTypeFromFact pred termList =
    let fType = List.map (fun x -> match x with
	| Int(_) -> Int(0)
	| String(_) -> String("")
	| Var(_) -> errNoVars (getLemmaString pred termList)
    ) termList
    in
	Hashtbl.add typeTable pred fType
;;

let checkTypeOfFact pred termList expType =
    if (List.length termList <> List.length expType) then errArity (getLemmaString pred termList);
    List.iter2 (fun element typeElem ->
	match element with
	| Int(_) -> (match typeElem with | Int(_) -> () | Var(_) -> () | _ ->
			errIntWhereString (getLemmaString pred termList))
	| String(_) -> (match typeElem with | String(_) -> () | Var(_) -> () | _ ->
			errStringWhereInt (getLemmaString pred termList))
	| Var(_) -> errNoVars (getLemmaString pred termList)
    ) termList expType
;;

let typeCheckFact fact =
    match fact with
    | Fact(pred,termList) ->
	(try
		let fType = Hashtbl.find typeTable pred in
		checkTypeOfFact pred termList fType
	with Not_found ->
		createTypeFromFact pred termList)
    | _ -> (assert(false))	(* this should never happen *)
;;

let inferTypesFromLits rule predList = 
    let typeHash = Hashtbl.create 5 in

    let matchOneLiteral pred termList eType =
	if (List.length termList <> List.length eType) then errArity (getRuleName rule);
	let changed = ref false in
	let newType = List.map2 (fun term expect -> match term with
	    | Var(x) ->
		(try
		    let typeSoFar = Hashtbl.find typeHash x in
		    match typeSoFar with
		    | Int(_) -> (match expect with
			| Int(_) -> expect
			| Var(_) -> changed := true; Int(0)
			| String(_) -> errMixup (getRuleName rule))
		    | String(_) -> (match expect with
			| String(_) -> expect
			| Var(_) -> changed := true; String("")
			| Int(_) -> errMixup (getRuleName rule))
		    | Var(_) ->
			Hashtbl.replace typeHash x expect; expect
		with Not_found ->
		     Hashtbl.add typeHash x expect;
		     expect)
	    | Int(_) ->	(match expect with
		| Int(_) -> expect
		| Var(_) -> changed := true; Int(0)
		| String(_) -> errIntWhereString (getRuleName rule)
	      )
	    | String(_) -> (match expect with
		| String(_) -> expect
		| Var(_) -> changed := true; String("")
		| Int(_) -> errStringWhereInt (getRuleName rule)
	      )
	) termList eType in

	if !changed then
		Hashtbl.replace typeTable pred newType
    in

    List.iter (fun (pred,termList) ->
	try
		let eType = Hashtbl.find typeTable pred in
		matchOneLiteral pred termList eType
	with Not_found ->
		let varType = List.map (fun x -> Var("x")) termList in
		Hashtbl.add typeTable pred varType;
		matchOneLiteral pred termList varType
    ) predList;
    typeHash
;;

let inferTypesFromFunctions rule typeHash functions =
    let rec checkIntExpression expr =
	match expr with
	| Plus(e1,e2) -> checkIntExpression e1; checkIntExpression e2
	| Minus(e1,e2) -> checkIntExpression e1; checkIntExpression e2
	| Terminal(t) -> match t with
		| Int(_) -> ()
		| String(_) -> errStringWhereInt (getRuleName rule)
		| Var(x) -> 
		    let typeSoFar = Hashtbl.find typeHash x in
		    match typeSoFar with
		    | Int(_) -> ()
		    | Var(_) -> Hashtbl.replace typeHash x (Int(0))
		    | String(_) -> errMixup (getRuleName rule)
    in

    (* provide for constant string assignments *)
    let checkExpression expr =
	match expr with
	| Terminal(t) -> (match t with
		| String(_) -> String("")
		| _ -> checkIntExpression expr; Int(0))
	| _ -> checkIntExpression expr; Int(0)
    in

    List.iter ( fun fn -> match fn with | FnEqual(var,expr) ->
	let typ = checkExpression expr in
	match var with
	| Var(x) -> Hashtbl.add typeHash x typ
	| _ -> (assert(false))	(* this should never happen *)
    ) functions
;;

let inferTypesFromFilters rule typeHash filters =
    let checkTerms t1 t2 =
	match t1 with
	| Int(_) -> (match t2 with
			| Int(_) -> ()
			| String(_) -> errComparison (getRuleName rule)
			| Var(x) -> let typeSoFar = Hashtbl.find typeHash x in
				match typeSoFar with
				| Int(_) -> ()
				| Var(_) -> Hashtbl.replace typeHash x (Int(0))
				| String(_) -> errComparison (getRuleName rule)
		    )
	| String(_) -> (match t2 with
			| String(_) -> ()
			| Int(_) -> errComparison (getRuleName rule)
			| Var(x) -> let typeSoFar = Hashtbl.find typeHash x in
				match typeSoFar with
				| String(_) -> ()
				| Var(_) -> Hashtbl.replace typeHash x (String(""))
				| Int(_) -> errComparison (getRuleName rule)
		    )
	| Var(x) -> let typeX = Hashtbl.find typeHash x in match t2 with
		| Int(_) -> (match typeX with
				| Int(_) -> ()
				| String(_) -> errComparison (getRuleName rule)
				| Var(_) -> Hashtbl.replace typeHash x (Int(0)))
		| String(_) -> (match typeX with
				| String(_) -> ()
				| Int(_) -> errComparison (getRuleName rule)
				| Var(_) -> Hashtbl.replace typeHash x (String("")))
		| Var(y) -> (let typeY = Hashtbl.find typeHash y in
			match typeX with
			| Int(_) -> (match typeY with
				| Int(_) -> ()
				| String(_) -> errComparison (getRuleName rule)
				| Var(_) -> Hashtbl.replace typeHash y (Int(0)))
			| String(_) -> (match typeY with
				| String(_) -> ()
				| Int(_) -> errComparison (getRuleName rule)
				| Var(_) -> Hashtbl.replace typeHash y (String("")))
			| Var(_) -> (match typeY with
				| String(_) -> Hashtbl.replace typeHash x (String(""))
				| Int(_) -> Hashtbl.replace typeHash x (Int(0))
				| Var(_) -> ())
		  )
    in

    List.iter (fun filter -> match filter with
	| LessThan(t1,t2)      -> checkTerms t1 t2
	| LessThanEqual(t1,t2) -> checkTerms t1 t2
	| GreaterThan(t1,t2)   -> checkTerms t1 t2
	| FilterEqual(t1,t2)   -> checkTerms t1 t2
    ) filters
;;

let typeCheckRule rule =
    match rule with
    | Rule(lhsPred,rhsPreds,negPreds,filters,functions) -> (
	let typeHash = inferTypesFromLits rule (List.append rhsPreds negPreds) in
	inferTypesFromFunctions rule typeHash functions;
	inferTypesFromFilters rule typeHash filters;

	let (pred,terms) = lhsPred in
	let newType = List.map (fun v -> 
		match v with
		| Var(x) -> Hashtbl.find typeHash x
	) terms in

	try
	    let oldType = Hashtbl.find typeTable pred in
	    if (List.length newType <> List.length oldType) then errArity (getRuleName rule);
	    let updatedType = List.map2 (fun oldt newt ->
		match oldt with
		| Var(_) -> newt
		| Int(_) -> ( match newt with
				| Int(_) -> newt
				| Var(_) -> oldt
				| String(_) -> errMixup (getRuleName rule)
			    )
		| String(_) -> ( match newt with
				| String(_) -> newt
				| Var(_) -> oldt
				| Int(_) -> errMixup (getRuleName rule)
			    )
	    ) oldType newType in
	    Hashtbl.replace typeTable pred updatedType

	with Not_found ->
	    Hashtbl.replace typeTable pred newType
        )
    | _ -> (assert(false))	(* this should never happen *)
;;