
open Ast
open Transformparseoutput

(*
	Translates a real-time well formed program datastructure into plain datalog rules.
	In particular, all temporal constructs like NEW AFTER NEXT will be translated into
	plain datalog rules, and logical construct OR will also be translated to plain
	datalog rules.
	
	--Find all unary edb predicates across all the clauses.
	--For each New Clause
		-- reachable(x) :- B1(x) B2(x) ~B3(x) ~B4(x).
	--For each next clause
		-- Find the unique variable on lhs. Call this x.
		-- Rename x on rhs to xold.
		-- Negate all those predicates which have xold in them
		-- Add LHS to RHS
		-- Add reachable(xold) to RHS
		-- reachable(x) is the LHS
		-- Add one Same(edb) for each edb that doesn't occur in the rule.
	--For each
*)		

(* This function will convert all Clauses including new, next, and queries,
   to regular datalog rules. A separate function will pretty print it as the translation 
 
let translateallRules (cList: Clause List) (stream: StreamWriter) = 
(
 let rulesList = ref [] in  
 List.iter ( fun x -> match x with 
            // Printing Edbs as is..
            | Edb(a) ->   Edb(a) (stream)
            | Query(a) -> translateQuery (a) (stream)
            | Rule(a,b,c) -> translateRule (a, b, c, stream)  
            | _ -> ()
           ) cList
) *)



let translateNewRule((newRule:Clause) , (edbPredicates:Set<Predicate>) ) =
(
	match newRule with 
	|NewRule(pupL, _) -> (
			let reachablePredicate = Pred("Reachable") in
			let head = List.hd pupL in
			match head with PosUp(BaseUnary(_,x)) -> (
				let lhs = PosLit(Base(reachablePredicate,x::[])) in
				let posListref = ref [] in
				let negListref = ref [] in
				(
					List.iter (
						fun y -> (
							match y with PosUp(BaseUnary(pred,term))
							-> (
								posListref := List.append (!posListref) (PosLit(Base(pred,x::[])) :: [])
							)
						)
					) pupL;
					
					Set.iter (
						fun z -> (
							let isPositive = List.exists (
								fun y -> match y with PosUp(BaseUnary(pred,_)) -> pred = z
							) pupL in
							(
								if isPositive then ()
								else (
									negListref := List.append (!negListref) (NegLit(Base(z,x::[])) :: [])
								)
							)
						)
					) edbPredicates;
					Rule(lhs, !posListref , !negListref);
				)
			)
		)
	
	|_ -> raise (new System.Exception("Cannot translate non-new Rule...") )
)

let replaceTermNameInTerm ((term:Term) , (oldName) , (newName)) =
(
	match term with Var(s) ->(
		if( s = oldName) then Var(newName)
		else term
	)
)

let replaceTermNamesInTermList ((tList:Term List) , oldName , newName) =
(
	let ret = ref[] in (
		List.iter(
			fun x -> ret := List.append (!ret) (replaceTermNameInTerm(x,oldName,newName)::[])
		) tList
	);
	!ret
)
let replaceTermNamesInBaseUnaryPredicate ((bup:BaseUnaryPredicate) ,(oldName) ,(newName))=
(
	match bup with BaseUnary(pred,term) -> BaseUnary(pred, replaceTermNameInTerm(term,oldName,newName));
)

let replaceTermNamesInPosUnaryPredicate ((pup:PositiveUnaryPredicate),oldName,newName)=
(
	match pup with PosUp(bup) -> PosUp(replaceTermNamesInBaseUnaryPredicate(bup,oldName,newName));
)
let replaceTermNamesInPosUPList( (posUPList:PositiveUnaryPredicate List),oldName,newName)=
(
	let ret = ref[] in(
		List.iter(
			fun x -> ret := List.append (!ret) (replaceTermNamesInPosUnaryPredicate(x,oldName,newName)::[])
		) posUPList
	);
	!ret
)

let replaceTermNamesInNegUnaryPredicate ((nup:NegativeUnaryPredicate),oldName,newName)=
(
	match nup with NegUp(bup) -> NegUp(replaceTermNamesInBaseUnaryPredicate(bup,oldName,newName));
)
let replaceTermNamesInNegUPList( (negUPList:NegativeUnaryPredicate List),oldName,newName)=
(
	let ret = ref[] in(
		List.iter(
			fun x -> ret := List.append (!ret) (replaceTermNamesInNegUnaryPredicate(x,oldName,newName)::[])
		) negUPList
	);
	!ret
)

let replaceTermNamesInBaseLiteral ((baseLiteral:BaseLiteral),oldName,newName ) =
(
	match baseLiteral with Base(pred,tList) -> Base(pred, replaceTermNamesInTermList(tList,oldName,newName))
)

let replaceTermNamesInPositiveLiteral ((posLit:PositiveLiteral),oldName,newName) =
(
	match posLit with PosLit(baseLit) -> PosLit(replaceTermNamesInBaseLiteral(baseLit,oldName,newName));
)
let replaceTermNamesInPosLitList ((posLitList:PositiveLiteral List),oldName,newName)=
(
	let ret = ref [] in(
		List.iter (
			fun x -> ret := List.append (!ret) (replaceTermNamesInPositiveLiteral(x,oldName,newName)::[])
		) posLitList
	);
	!ret
)

let replaceTermNamesInNegativeLiteral ((negLit:NegativeLiteral),oldName,newName) =
(
	match negLit with NegLit(baseLit) -> NegLit(replaceTermNamesInBaseLiteral(baseLit,oldName,newName));
)
let replaceTermNamesInNegLitList ((negLitList:NegativeLiteral List),oldName,newName)=
(
	let ret = ref [] in(
		List.iter (
			fun x -> ret := List.append (!ret) (replaceTermNamesInNegativeLiteral(x,oldName,newName)::[])
		) negLitList
	);
	!ret
)

(* next F(x) ~G(x) .. :- R(x_old) R(x_new)... *)
(* The x will be changed to x0. If x0 is already present
   in the list of variables, then x1 and so on.. *)
let getNewNameForXInNextRule ( tList , originalName) =
(
	let variables = Set.of_list tList in
	let ret = ref "" in
	let counter = ref 0 in (
		while( !counter  > -1) do
			if Set.mem (Var(originalName^string_of_int(!counter))) variables = false then (
				ret := originalName ^ string_of_int(!counter) ;
				counter := -1;
			)
			else
			counter := !counter + 1;
		done;
	);
	!ret;
)

let getLHSVariableNameInNextRule(pupList, nupList) =
(
	if (List.length pupList = 0) then (
		if(List.length nupList = 0) then (
			raise ( new System.Exception("Next rule with empty pup List and nup List .") )
		)
		else (
			let head = List.hd nupList in (
				match head with NegUp(BaseUnary(pred,Var(x))) -> x;
			)
		)
	)
	else (
		let head = List.hd pupList in (
			match head with PosUp(BaseUnary(pred,Var(x))) -> x;
		)
	)
)

let getSamePred (pred, x, xnew) = 
(
  let predname = match pred with 
                 |Pred(p) -> "Same"^p in
  let termList = List.append [Var(x)] [Var(xnew)] in                 
  PosLit(Base(Pred(predname), termList))
)

//Transform EDB predicates to Literal List
let transformedbPtoLitList (edbP, x) =
( 
  let litlist = ref [] in 
    Set.iter ( fun y -> match y with 
               | Pred(p) -> litlist := Base(Pred(p),[Var(x)])::!litlist
     ) edbP;
  !litlist   
)

let getNewPosLitVariableTermList (newposLit) = 
(
  match newposLit with 
  | PosLit(Base(p,q)) -> q
)

let replacexnewx ((x:string), (xnew:string), tList) =
(
 let repltList = ref [] in 
 List.iter (fun y -> match y with 
                    | Var(t) -> if (xnew.Equals(t:string)) then 
                                   repltList := Var(x)::(!repltList) 
                                   else 
                                   repltList := Var(t)::(!repltList)  
  ) tList;
  !repltList
)

let getBaseLiteralListfromPos (posLitList, x, xnew) = 
(
  let outlist = ref [] in 
  List.iter (fun y -> match y with
             | PosLit(Base(p,q)) ->  outlist := Base(p,replacexnewx(x, xnew, q)) :: (!outlist)
             ) posLitList;
  !outlist           
)

let getBaseLiteralListfromNeg (negLitList, x, xnew) = 
(
  let outlist = ref [] in 
  List.iter (fun y -> match y with
             | NegLit(Base(p,q)) -> outlist := Base(p,replacexnewx(x, xnew, q)) :: (!outlist)
             ) negLitList;
  !outlist           
)

let getSamePredLiterals((outpLlist), (outnLlist), (edbP), x, xnew) =
(
  let edblitlist = transformedbPtoLitList(edbP, x) in
  let edblitlistSet = Set.of_list(edblitlist) in   
  let opSet = Set.of_list(getBaseLiteralListfromPos (outpLlist, x, xnew)) in 
  let onSet = Set.of_list(getBaseLiteralListfromNeg (outnLlist, x, xnew)) in
  let unionS = Set.Union (opSet, onSet) in 
  let diffS = Set.Difference (edblitlistSet, unionS) in 
  let dList = Set.to_list (diffS) in 
  let sameList = ref [] in
  List.iter ( fun y ->
              match y with
              | Base(p,t) -> sameList := getSamePred(p, x, xnew):: !sameList ;               
             ) dList;
  !sameList
)

let getSameRules (pred, termList) =
(
 let rlist = ref [] in
 let samepredname = match pred with
                 | Pred(p) -> p in 
 let strlen = (String.length(samepredname)) in               
 let predname = String.sub samepredname 4 (strlen-4) in
 let tlist1 = [List.hd(termList)] in
 let tlist2 = List.tl(termList) in
 let poslit = PosLit(Base(pred,termList)) in 
 let poslitList = List.append ([PosLit(Base(Pred(predname),tlist1))]) ([PosLit(Base(Pred(predname), tlist2))]) in 
 let neglitList = List.append ([NegLit(Base(Pred(predname),tlist1))]) ([NegLit(Base(Pred(predname), tlist2))]) in
 let trueList = List.append ([PosLit(Base(Pred("True"), tlist1))]) ([PosLit(Base(Pred("True"), tlist2))]) in   
   rlist := (Rule(poslit,poslitList, []))::!rlist;
   rlist := (Rule(poslit,trueList, neglitList))::!rlist;
   !rlist
)

let convertsamePredLitstoRules (predList) = 
(
  let ruleList = ref [] in
    List.iter (fun x -> match x with
               | PosLit(Base(p,t)) -> ruleList := List.append (getSameRules(p,t)) (!ruleList)
               ) predList;
       !ruleList        
)


let translateNextRule ((clause:Clause ), (edbPredicates:Set<Predicate>)) =
(
	match clause with
	|NextRule(pupList,nupList,posLitList,negLitList) -> (
			let variables = ref [] in(
				variables := List.append (!variables) (Transformparseoutput.getVariablesList(pupList));
				variables := List.append (!variables) (Transformparseoutput.getVariablesList(nupList));
				variables := List.append (!variables) (Transformparseoutput.getVariablesList(posLitList));
				variables := List.append (!variables) (Transformparseoutput.getVariablesList(negLitList));
			);
			
			let x = getLHSVariableNameInNextRule(pupList, nupList) in
			let xnew = getNewNameForXInNextRule(!variables , x) in
			let reachablePredicate = Pred("Reachable") in
			let xReachable = PosLit(Base(reachablePredicate,Var(x)::[])) in
			let xnewReachable = PosLit(Base(reachablePredicate,Var(xnew)::[])) in
			(
			) ;
			let out_litList = ref [] in
			(
				out_litList := List.append (!out_litList) (PosLiteral(xReachable)::[]);
//Iterate over the PosLitList and check if each of them is
//1) edbp with same variable or 2) edbp with different variable or
//3) non edbp.				
				List.iter (
					fun z -> (
						match z with PosLit(Base(Pred(pred),tList)) -> (
							if(Set.mem (Pred(pred)) edbPredicates) then (
							//edb predicate.
								match tList with
								|Var(var)::tail -> (
										if(var=x) then (
											//same variable as on LHS. Negate
											out_litList := List.append (!out_litList) (NegLiteral(NegLit(Base(Pred(pred),tList)))::[])
										) else (
											//different variable. Don't negate.
											out_litList := List.append (!out_litList) (PosLiteral(z)::[])
										)
								)
								|_ -> (
									//this cannot happen.
								)
							) else (
							//non edb predicate. Plain translation.
								out_litList := List.append (!out_litList) (PosLiteral(z)::[]);
							)
						)
					)
				) posLitList ;
//Iterate over the NegLitList and check if each of them is
//1) edbp with same variable or 2) edbp with different variable or
//3) non edbp.		
				List.iter (
					fun z -> (
						match z with NegLit(Base(Pred(pred),tList)) -> (
							if(Set.mem (Pred(pred)) edbPredicates) then (
							//edb predicate.
								match tList with
								|Var(var)::tail -> (
										if(var=x) then (
											//same variable as on LHS. Negate
											out_litList := List.append (!out_litList) (PosLiteral(PosLit(Base(Pred(pred),tList)))::[])
										) else (
											//different variable. Don't negate.
											out_litList := List.append (!out_litList) (NegLiteral(z)::[])
										)
								)
								|_ -> (
									//this cannot happen.
								)
							) else (
							//non edb predicate. Plain translation.
								out_litList := List.append (!out_litList) (NegLiteral(z)::[]);
							)
						)
					)
				) negLitList ;
//Iterate over PosUPList and NegUpList and add them to out_litList				
				let posFromLhs = replaceTermNamesInPosUPList(pupList, x , xnew) in
				let negFromLhs = replaceTermNamesInNegUPList(nupList, x, xnew) in
				(
					List.iter(
						fun z -> out_litList := List.append (!out_litList) (PosUnary(z)::[])
					) posFromLhs;
					List.iter (
						fun z -> out_litList := List.append (!out_litList) (NegUnary(z)::[])
					)negFromLhs;
				);
//At this point out_litList contains everything on RHS in Literal List form.
//out_litList might also contain Zeroary , PoitiveUnary and NegativeUnary predicates.
//Take care to cast them into PosLiteral/NegLiteral.
//xnewReachable contains positiveLiteral in Rule(positiveLiteral,...,...)
				let out_posLitList = ref [] in
				let out_negLitList = ref [] in (
					List.iter (
						fun x -> match x with
						|PosLiteral(z) -> ( out_posLitList := List.append (!out_posLitList) (z::[]) )
						|NegLiteral(z) -> ( out_negLitList := List.append (!out_negLitList) (z::[]) )
						|PosUnary(PosUp(BaseUnary(pred,term))) -> (
								let z = (PosLit(Base(pred,term::[]))) in
									out_posLitList := List.append (!out_posLitList) (z::[])
							)
						|NegUnary(NegUp(BaseUnary(pred,term))) -> (
								let z = (NegLit(Base(pred,term::[]))) in
									out_negLitList := List.append (!out_negLitList) (z::[])
							)
						|Zeroary(z) -> (
							)
					) (!out_litList)
				);
// Add SameLists at this point before returning the new reachable rule
                let samepredLits = getSamePredLiterals(!out_posLitList, !out_negLitList, edbPredicates, x, xnew) in 
                out_posLitList  := List.append (!out_posLitList) (List.rev(samepredLits)) ;
				List.append ([Rule(xnewReachable , !out_posLitList , !out_negLitList)]) (convertsamePredLitstoRules(samepredLits))
			)
		)
	|_ -> raise (new System.Exception("Cannot translate non-next Rule...") )
)

(*
let translateTemporalClauseToPlainClause ((clause:Clause ), (edbPredicates:Set<Predicate>)) =
(
	match clause with
	|NewRule (a,b) -> translateNewRule (clause , edbPredicates)
	|NextRule(pupList,nupList,posLitList,negLitList) -> 
	  translateNextRule(clause, edbPredicates) 
	|_ -> clause
) 
*)

//Flatten query, collect all EDB predicates that the query depends upon
//For each variable in the query, find the EDB predicate it binds to and add a 
//reachable predicate for that variable in that rule 

(*
let processRuleforQuery(querypl, poslit, poslitlist, neglitlist) = 
(
 let 
 let querypred = match querypl with 
                 |PosLit(Base((Pred(pr)),t)) -> pr in 
 let (rulePred,varsList) = match p with
                           | PosLit(Base((Pred(p)),t)) -> (p, t) in 
 let                                             
             

)
*)

(*
let translateBaseQuerytoRules(querypl, clist) =
(
  let ruleList = ref [] in
  List.iter ( fun clause ->  match clause with
                             | Rule(p,pl,nl) -> ruleList := List.append (processRuleforQuery(querypl, p, pl, nl)) (!ruleList)
                             | _ -> ()
  ) clist;
  !ruleList
)
*)

(*
let translateQuery((query), (clist), (edbPreds)) = 
(
let rulelist = ref [] in  
  match query with
  | BaseQuery(pl) -> rulelist := List.append (translateBaseQuery(pl, clist)) (!rulelist);
  | ORQuery(q1,q2) -> rulelist := List.append (translateORQuery(q1, q2, clist)) (!rulelist);
  | AfterQuery(q1,q2) -> rulelist := List.append (translateAfterQuery(q1, q2, clist)) (!rulelist);
  (!rulelist)
)
*)


let translateTempClausesToRules((cList : Clause List), (edbPredicates: Set<Predicate>)) = 
(
  let tlist = ref [] in
  (*
  (
    List.iter (
         fun x -> ( match x with 
                   | NewRule(a,b) -> tlist := ( translateNewRule (x, edbPredicates)) :: (!tlist) 
                   | NextRule(puplist, nuplist, poslitlist, neglitlist) -> tlist := List.append (translateNextRule(x, edbPredicates)) (!tlist) 
                   | Query(q) -> tlist := List.append (translateQuery(q, cList)) (!tlist)
                   | _ -> ()
                  )
              ) cList; 
  )
  *)
  !tlist
) 


(*
let translateTemporalRulesToPlainRules ((cList : Clause List) , (edbPredicates:Set<Predicate>) ) =
(
	let ret = ref [] in
	(
		List.iter (
			fun x -> (
				let y = translateTemporalClauseToPlainClause (x , edbPredicates) in
				ret := List.append (!ret) (y::[])
			)
		) cList;
	);
	!ret
)
*)